from discord.abc import GuildChannel
from discord import TextChannel
from distest.exceptions import ResponseDidNotMatchError


async def assert_guild_channel_created(self, channel_name, timeout=None):
    """ Assert that the next channel created matches the name given

    :param str channel_name: The name of the channel to check for
    :param float timeout: The num of seconds to wait, defaults to the timeout provided at the start
    :returns: The channel that was created
    :rtype: discord.abc.GuildChannel
    :raises: NoResponseError
    """

    def check_for_channel_name(channel):
        return channel.name == channel_name

    return await self.wait_for_event(
        "guild_channel_create", check=check_for_channel_name, timeout=timeout
    )


async def assert_guild_channel_deleted(self, channel_name, timeout=None):
    """ Assert that the next channel deleted matches the name given

    TODO: check what the deleted channel actually returns

    :param str channel_name: The name of the channel to check for
    :param float timeout: The num of seconds to wait, defaults to the timeout provided at the start
    :returns: The channel that was deleted
    :rtype: discord.abc.GuildChannel
    :raises: NoResponseError
    """

    def check_for_channel_name(channel):
        return channel.name == channel_name

    return await self.wait_for_event(
        "guild_channel_delete", check=check_for_channel_name, timeout=timeout
    )


#
#
# async def assert_guild_channel_pin_content_equals(self, channel, message):
#     """ Checks when a channel has a message pinned, and that it matches the message provided. Doesn't work that well currently, maybe don't use
#
#     Useful to check if a specific message has been pinned.
#
#     For various reasons, this cannot currently check if the message was pinned or unpinned,
#         so please be careful to avoid any issues with mis-triggers. This may change in the future.
#
#     :param discord.TextChannel channel: The text channel to watch for pin changes on.
#     :param discord.Message message: The message to check the most recent pin against, compares IDs
#     """
#
#     # TODO: Check if I would use pins[0] or pins[-1] to get the most recent item
#
#     async def check(updated_channel, last_pin):
#         return updated_channel.id == channel.id
#
#     updated_channel, last_pin = await self.wait_for_event(
#         "guild_channel_pins_update", check=check, timeout=30
#     )
#
#     pins = await updated_channel.pins()
#
#     if pins[0].id == message.id:
#         return [updated_channel, last_pin]
#     else:
#         raise ResponseDidNotMatchError
#
#
# async def assert_guild_channel_unpin_content_equals(self, channel, message):
#     """ Checks when a channel's pin is updated.  Doesn't work at all rn, maybe don't use
#
#     """
#
#     def check(updated_channel,):
#         return updated_channel.id == channel.id
#
#     return



"""
The TestCollector Class and some supporting code.

Each test function in the tester bot should be decorated with an instance of TestCollector(),
and must have a unique name. The TestCollector() is then passed onto the bot, which runs the tests.
"""

from .TestInterface import Test


class ExpectCalls:
    """ Wrap a function in an object which counts the number of times it was called. If the number
    of calls is not equal to the expected number when this object is garbage collected, something has
    gone wrong, and in that case an error is thrown.

    :param function function: The test :py:class:`function` to track
    :param int expected_calls: The number of calls expected for that function. Defaults to 1,
                               and is not currently able to be set to another value
    """

    def __init__(self, function, expected_calls=1):
        self.function = function
        self.expected_calls = expected_calls
        self.call_count = 0

    def __call__(self, *args, **kwargs):
        """ Increment ``call_count`` when the function is called, then actually call the function. """
        self.call_count += 1
        return self.function(*args, **kwargs)

    def __del__(self):
        """ Create an error that contains the call information if the call count is wrong when the test is deleted"""
        if self.call_count != self.expected_calls:
            message = (
                "{} was called {} times. It was expected to have been called {} times"
            )
            raise RuntimeError(
                message.format(self.function, self.call_count, self.expected_calls)
            )


class TestCollector:
    """ Used to group tests and pass them around all at once.

    Tests can be either added with :func:`add <distest.collector.TestCollector.add>` or by using ``@TestCollector``
    to decorate the function, as seen in the sample code below. Is very similar in function to
    :py:class:`Command <discord.ext.commands.Command>` from discord.py, which you might already be familiar with.

    .. literalinclude:: ../../../example_tester.py
       :linenos:
       :language: python3
       :lines: 38-50
       :emphasize-lines: 1, 8

    """

    def __init__(self):
        self._tests = []

    def add(self, function, name=None, needs_human=False):
        """ Adds a test function to the group, if one with that name is not already present

        :param func function: The function to add
        :param str name: The name of the function to add, defaults to the function name but can be overridden
                         with the provided name just like with :py:class:`discord.ext.commands.Command`.
                         See sample code above.
        :param bool needs_human: Optional boolean, true if the test requires a human interaction
        """
        name = name or function.__name__
        test = Test(name, function, needs_human=needs_human)
        if name in self._tests:
            raise KeyError("A test case called {} already exists.".format(name))
        self._tests.append(test)

    def find_by_name(self, name):
        """ Return the test with the given name, return ``None`` if it does not exist.

        :param str name: The name of the test
        """
        for i in self._tests:
            if i.name == name:
                return i
        return None

    def __call__(self, *args, **kwargs):
        """ Add a test decorator-style, simply calls `add` when used to decorate something. """

        def _decorator(function):
            self.add(function, *args, **kwargs)

        return ExpectCalls(_decorator, 1)

    def __iter__(self):
        """ Makes the `TestCollector` able to be iterated over, which is really helpful in a number of cases."""
        return (i for i in self._tests)



async def send_message(self, content):
    """ Send a message to the channel the test is being run in. **Helper Function**

    :param str content: Text to send in the message
    :returns: The message that was sent
    :rtype: discord.Message
    """
    return await self.channel.send(content)


def _check_message(self, message):
    return message.channel == self.channel and message.author == self.target


async def edit_message(message, new_content):
    """ Modify a message. Most tests and ``send_message`` return the ``discord.Message`` they sent, which can be
    used here. **Helper Function**

    :param discord.Message message: The target message. Must be a ``discord.Message``
    :param str new_content: The text to change `message` to.
    :returns: `message` after modification.
    :rtype: discord.Message
    """
    return await message.edit(content=new_content)



from discord import Message, Embed

from distest.exceptions import ResponseDidNotMatchError


async def assert_embed_equals(
    message: Message, matches: Embed, attributes_to_prove: list = None,
):
    """If ``matches`` doesn't match the embed of ``message``, fail the test.

    Checks only the attributes from ``attributes_to_prove``.

    :param message: original message
    :param matches: embed object to compare to
    :param attributes_to_prove: a string list with the attributes of the embed, which are to compare
        This are all the Attributes you can prove: "title", "description", "url", "color",
        "author", "video", "image" and "thumbnail".
    :return: message
    :rtype: discord.Message
    """

    # All possible attributes a user can set during initialisation
    possible_attributes = [
        "title",
        "description",
        "url",
        "color",
        "author",  # This is not the original author of the message, author is a attribute you are able to set.
        "video",
        "image",
        "thumbnail",
    ]
    # View all (visible) attributes visualized here: https://imgur.com/a/tD7Ibc4

    attributes = []

    # Proves, if the attribute provided by the user is a valid attribute to check
    if attributes_to_prove is not None:
        for value in attributes_to_prove:
            if value not in possible_attributes:
                raise NotImplementedError('"' + value + '" is not a possible value.')
            attributes.append(value)
    else:
        # If no attributes to check are provided, check them all.
        attributes = possible_attributes

    for embed in message.embeds:
        for attribute in attributes:
            if attribute == "image" or attribute == "thumbnail":
                # Comparison of Embedded Images / Thumbnails
                if getattr(getattr(embed, attribute), "url") != getattr(
                    getattr(matches, attribute), "url"
                ):
                    raise ResponseDidNotMatchError(
                        "The {} attribute did't match".format(attribute)
                    )
            elif attribute == "video":
                # Comparison of Embedded Video
                if getattr(getattr(embed, "video"), "url") != getattr(
                    getattr(matches, "video"), "url"
                ):
                    raise ResponseDidNotMatchError("The video attribute did't match")
            elif attribute == "author":
                # Comparison of Author
                if getattr(getattr(embed, "author"), "name") != getattr(
                    getattr(matches, "author"), "name"
                ):
                    raise ResponseDidNotMatchError("The author attribute did't match")
            elif not getattr(embed, attribute) == getattr(matches, attribute):
                print(
                    "Did not match:",
                    attribute,
                    getattr(embed, attribute),
                    getattr(matches, attribute),
                )
                raise ResponseDidNotMatchError
    return message



"""
Stores all the Exceptions that can be called during testing.

Allows for a more through understanding of what went wrong. Not all of these are currently in use.
"""


class TestRequirementFailure(Exception):
    """ Base class for the special errors that are raised when an expectation is not met during testing """


class NoResponseError(TestRequirementFailure):
    """ Raised when the target bot fails to respond to a message """


class NoReactionError(TestRequirementFailure):
    """ Raised when the target bot failed to react to a message """


class UnexpectedResponseError(TestRequirementFailure):
    """ Raised when the target bot failed to stay silent """


class ErrordResponseError(TestRequirementFailure):
    """ Raised when the target bot produced an error message """


class UnexpectedSuccessError(TestRequirementFailure):
    """ Raised when the target bot failed to produce an error message """


class HumanResponseTimeout(TestRequirementFailure):
    """ Raised when a human fails to assert the result of a test """


class HumanResponseFailure(TestRequirementFailure):
    """ Raised when a human fails a test """


class ResponseDidNotMatchError(TestRequirementFailure):
    """ Raised when the target bot responds with a message that doesn't meet criteria """


class ReactionDidNotMatchError(TestRequirementFailure):
    """ Raised when the target bot reacts with the wrong emoji """



from distest.exceptions import ResponseDidNotMatchError, UnexpectedResponseError
from re import match


async def assert_message_equals(message, matches):
    """ If ``message`` does not match a string exactly, fail the test.

    :param discord.Message message: The message to test.
    :param str matches: The string to test `message` against.
    :returns: `message`
    :rtype: discord.Message
    :raises: ResponseDidNotMatchError
    """
    if message.content != matches:
        raise ResponseDidNotMatchError
    return message


async def assert_message_contains(message, substring):
    """ If `message` does not contain the given substring, fail the test.

    :param discord.Message message: The message to test.
    :param str substring: The string to test `message` against.
    :returns: `message`
    :rtype: discord.Message
    :raises: ResponseDidNotMatchError
    """
    if substring not in message.content:
        raise ResponseDidNotMatchError
    return message


async def assert_message_matches(message, regex):
    """ If `message` does not match a regex, fail the test.

    Requires a properly formatted Python regex ready to be used in the ``re`` functions.


    :param discord.Message message: The message to test.
    :param str regex: The regular expression to test `message` against.
    :returns: `message`
    :rtype: discord.Message
    :raises: ResponseDidNotMatchError
    """
    if not match(regex, message.content):
        raise ResponseDidNotMatchError
    return message


async def assert_message_has_image(message):
    """ Assert `message` has an attachment. If not, fail the test.

    :param discord.Message message: The message to test.
    :returns: `message`
    :rtype: discord.Message
    :raises: UnexpectedResponseError
    """
    if message.attachments == [] and message.embeds == []:
        raise UnexpectedResponseError
    return message



import re
import argparse


def token_arg(token_value: str) -> str:
    """
    Validates the discord token, won't run the bot if the token is not valid.

    Matches to the regex: [\w\d]{24}\.[\w\d]{6}\.[\w\d-_]{27}.
    Thanks to https://github.com/AnIdiotsGuide/Tutorial-Bot/blob/Episode-4/app.js for the regex I used.
    Thanks to https://gist.github.com/mikecharles/9ed3082b10d77d658743 for an example of how this works with argparse

    :param token_value: The string that may be a bot token
    :return token_value: Returns the token_value if the token matches the regex
    :exception ArgumentParseException: Raises the exception if the token_value does not match the regex
    """

    # Regex check
    if not re.match(r"[MN][A-Za-z\d]{23}\.[\w-]{6}\.[\w-]{27}", token_value):
        raise argparse.ArgumentTypeError(
            "must be a correct Discord bot token.".format(token_value)
        )

    # If that passes, the token's value will be returned.
    return token_value



from distest.exceptions import ReactionDidNotMatchError


async def assert_reaction_equals(self, contents, emoji):
    """ Send a message and ensure that the reaction is equal to `emoji`. If not, fail the test.

    :param str contents: The content of the trigger message. (A command)
    :param discord.Emoji emoji: The emoji that the reaction must equal.
    :returns: The resultant reaction object.
    :rtype: discord.Reaction
    :raises: ReactionDidNotMatchError
    """
    reaction = await self.wait_for_reaction(await self.send_message(contents))
    if str(reaction[0].emoji) != emoji:
        raise ReactionDidNotMatchError
    return reaction



import argparse
import sys

from .bot import DiscordInteractiveInterface, DiscordCliInterface
from .collector import TestCollector
from .TestInterface import TestInterface


def run_dtest_bot(sysargs, test_collector, timeout=5):
    """ This is the function you will call in your test suite's ``if __name__ == "__main__":`` statement to
    get the bot started.

    :param list sysargs: The list returned by ``sys.argv``, this function parses it and will handle errors in format
    :param TestCollector test_collector: The :ref:`TestCollector` that has been used to decorate the tests
    :param int timeout: An optional parameter to override the amount of time to wait for responses before failing
                        tests. Defaults to 5 seconds.
    """
    from distest.validate_discord_token import token_arg

    all_run_options = ["all"]
    for i in test_collector._tests:
        all_run_options.append(i.name)

    parser = argparse.ArgumentParser(
        description="A small library used to write automated unit tests for Discord bots. "
                    "Has 2 modes, Interactive and CLI. "
                    "If you include -c, the bot will expect to be used in CLI mode. "
                    "See the github wiki for more info"
    )
    required = parser.add_argument_group("Always Required Arguments")
    required.add_argument(
        "bot_target",
        metavar="target_bot_user",
        type=int,
        nargs=1,
        help="The client ID of the target bot.",
    )
    required.add_argument(
        "bot_token",
        metavar="tester_bot_token",
        type=token_arg,
        nargs=1,
        help="The bot token for the testing bot (this bot).",
    )
    cli_only = parser.add_argument_group("CLI Only")
    cli_only.add_argument(
        "--channel",
        "-c",
        metavar="channel",
        type=int,
        nargs=1,
        help="The channel ID that the tests should be occurring in (CLI) "
             "or the ID to send the awake message to (Interactive)",
        dest="channel",
    )
    run_stats_group = cli_only.add_mutually_exclusive_group()
    run_stats_group.add_argument(
        "--run",
        "-r",
        type=str,
        choices=all_run_options,
        help="Runs the bot in run mode, equivalent to ::run <option>. "
             "Options are listed, they are the tests declared in the bot and the three default options."
             "Required for the bot to be run in CLI mode, if using in Interactive mode, don't specify this",
    )
    run_stats_group.add_argument(
        "--stats",
        "-s",
        action="store_true",
        help="Runs the bot in stats mode, outputting the last runs stats. "
             "Equivalent to ::stats",
    )
    parser.add_argument(
        "--timeout",
        "-t",
        type=int,
        nargs=1,
        help="Changes the timeout (in seconds) on tests before they are assumed to have failed. "
             "Default is 5 sec.",
    )

    sysargs.pop(0)  # Pops off the first arg (the filename that is being run)
    clean_args = vars(parser.parse_args(sysargs))

    # Makes the changing of the timeout optional
    if clean_args.get("timeout") is not None:
        timeout = clean_args.get("timeout")[0]

    # Controls whether or not the bot is run in CLI mode based on the parameters present
    if clean_args["run"] is not None:
        # If --run is present, the bot should be in CLI mode
        print("In CLI mode")
        run_command_line_bot(
            clean_args.get("bot_target")[0],
            clean_args.get("bot_token")[0],
            clean_args.get("run"),
            clean_args.get("channel")[0],
            clean_args.get("stats"),
            test_collector,
            timeout,
        )
    else:
        print("Not in CLI mode")
        run_interactive_bot(
            clean_args.get("bot_target")[0],
            clean_args.get("bot_token")[0],
            test_collector,
            timeout,
        )


def run_interactive_bot(target_name, token, test_collector, timeout=5):
    """ Run the bot in interactive mode.

        Relies on :py:func:`run_dtest_bot` to parse the command line arguments and pass them here.
        Not really meant to be called by the user.

        :param str target_name: The display name of the bot we are testing.
        :param str token: The tester's token, used to log in.
        :param TestCollector test_collector: The collector that gathered our tests.
        :param int timeout: The amount of time to wait for responses before failing tests.
    """

    bot = DiscordInteractiveInterface(target_name, test_collector, timeout)
    bot.run(token)  # Starts the bot


def run_command_line_bot(target, token, tests, channel_id, stats, collector, timeout):
    """ Start the bot in command-line mode. The program will exit 1 if any of the tests failed.

        Relies on :py:func:`run_dtest_bot` to parse the command line arguments and pass them here.
        Not really meant to be called by the user.

        :param str target: The display name of the bot we are testing.
        :param str token: The tester's token, used to log in.
        :param str tests: List of tests to run.
        :param int channel_id: The ID of the channel in which to run the tests.
        :param bool stats: Determines whether or not to display stats after run.
        :param TestCollector collector: The collector that gathered our tests.
        :param int timeout: The amount of time to wait for responses before failing tests.
    """
    bot = DiscordCliInterface(target, collector, tests, channel_id, stats, timeout)
    failed = bot.run(token)  # returns True if a test failed
    sys.exit(1 if failed else 0)  # Calls sys.exit based on the state of `failed`



from concurrent.futures import _base
from discord import Reaction
from distest.exceptions import (
    UnexpectedResponseError,
    HumanResponseTimeout,
    HumanResponseFailure,
)


async def ensure_silence(self):
    """ Assert that the bot does not post any messages for some number of seconds.

    :raises: UnexpectedResponseError, TimeoutError
    """
    try:
        await self.client.wait_for(
            "message", timeout=self.client.timeout, check=self._check_message
        )
    except _base.TimeoutError:
        pass
    else:
        raise UnexpectedResponseError


async def ask_human(self, query):
    """ Ask a human for an opinion on a question using reactions.

    Currently, only yes-no questions are supported. If the human answers 'no', the test will be failed. Do not use
    if avoidable, since this test is not really automateable. Will fail if the reaction is wrong or takes too long
    to arrive

    :param str query: The question for the human.
    :raises: HumanResponseTimeout, HumanResponseFailure
    """
    message = await self.send_message(query)
    await message.add_reaction("\u2714")
    await message.add_reaction("\u274C")

    def check(human_reaction, user):
        if human_reaction.count > 1:
            return human_reaction.message

    try:
        reaction: Reaction = await self.client.wait_for(
            "reaction_add", timeout=self.client.timeout, check=check
        )
    except _base.TimeoutError:
        raise HumanResponseTimeout
    else:
        reaction, _ = reaction
        if reaction.emoji == "\u274c":
            raise HumanResponseFailure



import enum


class TestResult(enum.Enum):
    """ Enum representing the result of running a test case """

    UNRUN = 0
    SUCCESS = 1
    FAILED = 2


SPECIAL_TEST_NAMES = {"all", "unrun", "failed"}


class Test:
    """ Holds data about a specific test.

    :param str name: The name of the test, checks this against the valid test names
    :param function func: The function in the tester bot that makes up this test
    :param bool needs_human: Weather or not this test will require human interaction to complete
    :raises: ValueError
    """

    def __init__(self, name, func, needs_human=False):
        if name in SPECIAL_TEST_NAMES:
            raise ValueError("{} is not a valid test name".format(name))
        self.name = name
        self.func = func
        self.last_run = 0
        self.result = TestResult.UNRUN
        self.needs_human = needs_human


class TestInterface:
    """ All the tests, and some supporting functions. Tests are designed to be run
    by the tester and mixed together in order to actually test the bot.

    .. note::
        In addition to the tests failing due to their own reasons, all tests will also fail if they timeout.
        This period is specified when the bot is run.

    .. note::
        Some functions (``send_message`` and ``edit_message``) are helper functions rather than tests and serve to bring
        some of the functionality of the discord library onto the same level as the tests.

    .. note::
        ``assert_reply_*`` tests will send a message with the passed content, while ``assert_message_*`` tests require a
        ``Message`` to be passed to them. This allows for more flexibility when you need it and an easier
        option when you don't.

    :param discord.Client client: The discord client of the tester.
    :param discord.TextChannel channel: The discord channel in which to run the tests.
    :param discord.Member target: The bot we're testing.
    """

    def __init__(self, client, channel, target):
        self.client = client
        self.channel = channel
        self.target = target
        self.voice_client = None
        self.voice_channel = None

    # Imported Methods
    from ._helpers import send_message, _check_message, edit_message
    from ._voice import connect, disconnect
    from ._oddballs import ask_human, ensure_silence
    from ._reaction import assert_reaction_equals
    from ._message import (
        assert_message_equals,
        assert_message_contains,
        assert_message_has_image,
        assert_message_matches,
    )
    from ._reply import (
        assert_reply_equals,
        assert_reply_contains,
        assert_reply_embed_equals,
        assert_reply_matches,
        assert_reply_has_image,
        get_delayed_reply,
    )
    from ._embeds import assert_embed_equals
    from ._wait_for import (
        wait_for_message,
        wait_for_reaction,
        wait_for_reply,
        wait_for_event,
        wait_for_message_in_channel,
    )
    from ._guild_channel import (
        assert_guild_channel_created,
        assert_guild_channel_deleted,
        # assert_guild_channel_pin_content_equals,
        # assert_guild_channel_unpin_content_equals,
    )

    edit_message = staticmethod(edit_message)
    assert_message_equals = staticmethod(assert_message_equals)
    assert_message_contains = staticmethod(assert_message_contains)
    assert_message_has_image = staticmethod(assert_message_has_image)
    assert_message_matches = staticmethod(assert_message_matches)
    assert_embed_equals = staticmethod(assert_embed_equals)



from distest.exceptions import NoResponseError
from concurrent.futures import _base
from typing import Callable, Optional


async def wait_for_reaction(self, message):
    """ Assert that ``message`` is reacted to with any reaction.

    :param discord.Message message: The message to test with
    :returns: The reaction object.
    :rtype: discord.Reaction
    :raises NoReactionError:
    """

    def check_reaction(reaction, user):
        return (
            reaction.message.id == message.id
            and user == self.target
            and reaction.message.channel == self.channel
        )

    try:
        result = await self.client.wait_for(
            "reaction_add", timeout=self.client.timeout, check=check_reaction
        )
    except _base.TimeoutError:
        raise NoResponseError
    else:
        return result


async def wait_for_message(self):
    """ Wait for the bot the send any message. Will fail on timeout, but will ignore messages sent by anything other
    that the target.

    :returns: The message we've been waiting for.
    :rtype: discord.Message
    :raises: NoResponseError
    """
    try:
        result = await self.client.wait_for(
            "message", timeout=self.client.timeout, check=self._check_message
        )
    except _base.TimeoutError:
        raise NoResponseError
    else:
        return result


async def wait_for_message_in_channel(self, content, channel_id):
    """ Send a message with ``content`` and returns the next message that the targeted bot sends. Used in many other
    tests.

    :param str content: The text of the trigger message.
    :param int channel_id: The id of the channel that the message is sent in.
    :returns: The message we've been waiting for.
    :rtype: discord.Message
    :raises: NoResponseError
    """

    def check_for_message_in_channel(message):
        return message.channel.id == channel_id and message.content == content

    return await self.wait_for_event(
        "message", check=check_for_message_in_channel, timeout=30
    )


async def wait_for_reply(self, content):
    """ Send a message with ``content`` and returns the next message that the targeted bot sends. Used in many other
    tests.

    :param str content: The text of the trigger message.
    :returns: The message we've been waiting for.
    :rtype: discord.Message
    :raises: NoResponseError
    """
    await self.channel.send(content)
    return await self.wait_for_message()


async def wait_for_event(
    self, event: str, check: Optional[Callable[..., bool]] = None, timeout: float = None
):
    """ A wrapper for the discord.py function :py:func:`wait_for <discord.Client.wait_for>`, tuned to be useful for distest.

    See https://discordpy.readthedocs.io/en/latest/api.html#event-reference for a list of events.

    :param event: The discord.py event, as a string and with the ``on_`` removed from the beginning.
    :param Callable[...,bool] check: A check function that all events of the type are ran against. Should return true when the desired event occurs, takes the event's params as it's params
    :param float timeout: How many seconds to wait for the event to occur.
    :return: The parameters of the event requested
    :raises: NoResponseError
    """
    if timeout is None:
        timeout = self.client.timeout

    try:
        result = await self.client.wait_for(event, timeout=timeout, check=check)
    except _base.TimeoutError:
        raise NoResponseError
    # TODO: What happens if the event is wrong / not valid?
    else:
        return result



#!/usr/bin/env python3
# encoding: utf-8

from setuptools import setup


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="distest",
    version="0.4.9",
    description="Automate the testing of discord bots... With discord bots!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/JakeCover/distest",
    author="Jake Cover",
    author_email="python@jakecover.me",
    license="MIT",
    packages=["distest", "distest.TestInterface"],
    install_requires=["discord.py>=1.5.0"],
    zip_safe=False,
    classifiers=["Topic :: Software Development :: Testing :: Unit"],
    keywords=[
        "Discord",
        "Discord.py",
        "Unit Test",
        "Test",
        "Distest",
        "Discord Testing",
    ],
)



async def connect(self, channel):
    """
    Connect to a given VoiceChannel
    :param channel: The VoiceChannel to connect to.
    :return:
    """
    self.voice_channel = self.client.get_channel(channel)
    self.voice_client = await self.voice_channel.connect()


async def disconnect(self):
    """
    Disconnect from the VoiceChannel; Doesn't work if the Bot isn't connected.
    :return:
    """
    if self.voice_channel is None:
        raise NotImplementedError("The Bot isn't connected.")
    await self.voice_client.disconnect()



""" Contains the discord clients used to run tests.

:py:class:`DiscordBot` contains the logic for running tests and finding the target bot

:py:class:`DiscordInteractiveInterface` is a subclass of :py:class:`DiscordBot` and contains the logic to handle
commands sent from discord to run tests, display stats, and more

:py:class:`DiscordCliInterface` is a subclass of :py:class:`DiscordInteractiveInterface` and simply contains logic to
start the bot when it wakes up
"""
import discord

from .TestInterface import TestResult, Test, TestInterface
from .exceptions import TestRequirementFailure
from .collector import TestCollector

HELP_TEXT = """\
**::help** - Show this help
**::run** all - Run all tests
**::run** unrun - Run all tests that have not been run
**::run** *name* - Run a specific test
**::list** - List all the tests and their status
"""

intents = discord.Intents.default()
intents.members = True


class DiscordBot(discord.Client):
    """ Discord bot used to run tests.
    This class by itself does not provide any useful methods for human interaction,
    and is just used as a superclass of the two interfaces,
    :py:class:`DiscordInteractiveInterface` and :py:class:`DiscordCliInterface`

    :param str target_id: The name of the target bot, used to ensure that the target user is actually
                            present in the server. Good for checking for typos or other simple mistakes.
    """

    def __init__(self, target_id):
        super().__init__(intents=intents)
        self._target_name = target_id

    def _find_target(self, server: discord.Guild) -> discord.Member:
        """ Confirms that the target user is actually present in the specified guild

            :param discord.Guild server: The ``discord.Guild()`` object of the guild
            to look for the target user in
            :rtype:  discord.Member
        """
        for member in server.members:
            if self._target_name == member.id:
                if member.status == discord.Status.offline:
                    print("Looks like the target bot is on the server but offline, you might want to check on that!")
                return member
        raise KeyError("Could not find member with id {}".format(self._target_name))

    async def run_test(
        self, test: Test, channel: discord.TextChannel, stop_error=False
    ) -> TestResult:
        """ Run a single test in a given channel.

            Updates the test with the result and returns it

            :param Test test: The :py:class:`Test` that is to be run
            :param discord.TextChannel channel: The
            :param stop_error: Weather or not to stop the program on error. Not currently in use.
            :return: Result of the test
            :rtype: TestResult
        """
        test_interface = TestInterface(self, channel, self._find_target(channel.guild))
        try:
            print("Running test: {}".format(test.name))
            await test.func(test_interface)
        except TestRequirementFailure:
            test.result = TestResult.FAILED
            if not stop_error:
                raise
        else:
            test.result = TestResult.SUCCESS
        return test.result


class DiscordInteractiveInterface(DiscordBot):
    """ A variant of the discord bot which commands sent in discord to allow
    a human to run the tests manually.

    Does NOT support CLI arguments

    :param str target_id: The name of the bot to target (Username, no discriminator)
    :param TestCollector collector: The instance of Test Collector that contains the tests to run
    :param int timeout: The amount of time to wait for responses before failing tests.
    """

    def __init__(self, target_id, collector: TestCollector, timeout=5):
        super().__init__(target_id)
        self._tests = collector
        self.timeout = timeout
        self.failure = False

    async def _run_by_predicate(self, channel, predicate=lambda test: True):
        """ Iterate through ``_tests`` and run any test for which ``predicate`` returns True

        :param discord.TextChannel channel: The channel to run the test in.
        :param function predicate: The check a test must pass to be run.
        """
        for test in self._tests:
            if predicate(test):
                await self.run_test(test, channel, stop_error=True)

    async def _build_stats(self, tests) -> str:
        """ Helper function for constructing the stat display based on test status.
        Iterate over each test in ``tests`` and creates a string (``response``) based
        on the result property of each ``Test``

        :param list[Test] tests: The list of tests used to create the stats
        :return: Ready-to-send string congaing the results of the tests, including
                 discord markdown
        :rtype: str
        """
        response = "```\n"
        longest_name = max(map(lambda t: len(t.name), tests))
        for test in tests:
            response += test.name.rjust(longest_name) + " "
            if test.needs_human:
                response += "✋ "
            else:
                response += "   "
            if test.result is TestResult.UNRUN:
                response += "⚫ Not run\n"
            elif test.result is TestResult.SUCCESS:
                response += "✔️ Passed\n"
            elif test.result is TestResult.FAILED:
                response += "❌ Failed\n"
                self.failure = True
        response += "```\n"
        return response

    async def _display_stats(self, channel: discord.TextChannel):
        """ Display the status of the various tests. Just a send wrapper for
        :py:func:`_build_stats`
        """
        await channel.send(await self._build_stats(self._tests))

    async def on_ready(self):
        """ Report when the bot is ready for use and report the available tests
        to the console
        """
        print("Started distest bot.")
        print(f"Invite Link: https://discordapp.com/oauth2/authorize?client_id={self.user.id}&scope=bot&permissions=8 ")
        print("Available tests are:")
        for test in self._tests:
            print("   {}".format(test.name))

    async def on_message(self, message: discord.Message):
        """ Handle an incoming message, see :py:func:`discord.event.on_message` for
        event reference.

        Parse a message, can ignore it or parse the message as a command and
        run some tests or do one of the alternate functions (stats, list, or help)

        :param discord.Message message: The message being recieved, passed by discord.py
        """
        if message.author == self.user:
            return
        if not isinstance(message.channel, (discord.DMChannel, discord.GroupChannel)):
            if message.content.startswith("::run "):
                name = message.content[6:]
                await self.run_tests(message.channel, name)
                await self._display_stats(message.channel)
            elif message.content in ["::stats", "::list"]:
                await self._display_stats(message.channel)
            elif message.content == "::help":
                await message.channel.send(HELP_TEXT)

    async def run_tests(self, channel: discord.TextChannel, name: str):
        """ Helper function for choosing and running an appropriate suite of tests
        Makes sure only tests that still need to be run are run, also prints
        to the console when a test is run

        :param discord.TextChannel channel: The channel in which to run the tests
        :param str name: Selector string used to determine what category of test to run
        """
        print("Running: ", name)
        if name == "all":
            await self._run_by_predicate(channel)
        elif name == "unrun":
            await self._run_by_predicate(
                channel, lambda test: test.result is TestResult.UNRUN
            )
        elif name == "failed":
            await self._run_by_predicate(
                channel, lambda test: test.result is TestResult.FAILED
            )
        elif self._tests.find_by_name(name) is None:
            text = ":x: There is no test called `{}`"
            await channel.send(channel, text.format(name))
        else:
            print("Running test: {}".format(name))
            await self.run_test(self._tests.find_by_name(name), channel)


class DiscordCliInterface(DiscordInteractiveInterface):
    """ A variant of the discord bot which is designed to be run off command line arguments.

    :param str target_id: The name of the bot to target (Username, no discriminator)
    :param TestCollector collector: The instance of Test Collector that contains the
                                    tests to run
    :param str test: The name of the test option (all, specific test, etc)
    :param int channel_id: The ID of the channel to run the bot in
    :param bool stats: If true, run in hstats mode.
    """

    def __init__(self, target_id, collector, test, channel_id, stats, timeout):
        super().__init__(target_id, collector, timeout)
        self._test_to_run = test
        self._channel_id = channel_id
        self._stats = stats
        self._channel = None

    #
    def run(self, token) -> int:
        """ Override of the default run() that returns failure state after completion.
        Allows the failure to cascade back up until it is processed into an exit code by
        :py:func:`run_command_line_bot`

        :param str token: The tester bot token
        :return: Returns 1 if the any test failed, otherwise returns zero.
        :rtype: int
        """
        super().run(token)
        return self.failure

    async def on_ready(self):
        """ Run all the tests sequentially when the bot becomes awake and exit when the
        tests finish. The CLI should run all by itself without prompting, and this allows it
        to behave that way.
        """
        self._channel = self.get_channel(self._channel_id)
        print("Started distest bot.")
        print(f"Invite Link: https://discordapp.com/oauth2/authorize?client_id={self.user.id}&scope=bot&permissions=8 ")
        if self._test_to_run is not None:
            await self.run_tests(self._channel, self._test_to_run)
            await self._display_stats(self._channel)
        elif self._stats:
            await self._display_stats(self._channel)
        await self.close()



from inspect import signature, _ParameterKind
from discord import Embed, Message
from asyncio import sleep


async def assert_reply_equals(self, contents, matches):
    """ Send a message and wait for a response. If the response does not match the string
    exactly, fail the test.

    :param str contents: The content of the trigger message. (A command)
    :param str matches: The string to test against.
    :returns: The reply.
    :rtype: discord.Message
    :raises: ResponseDidNotMatchError
    """
    response = await self.wait_for_reply(contents)
    return await self.assert_message_equals(response, matches)


async def assert_reply_contains(self, contents, substring):
    """ Send a message and wait for a response. If the response does not contain
    the given substring, fail the test.

    :param str contents: The content of the trigger message. (A command)
    :param str substring: The string to test against.
    :returns: The reply.
    :rtype: discord.Message
    :raises: ResponseDidNotMatchError
    """
    response = await self.wait_for_reply(contents)
    return await self.assert_message_contains(response, substring)


async def assert_reply_embed_equals(
    self, message: str, equals: Embed, attributes_to_check: list = None
):
    response = await self.wait_for_reply(message)
    return await self.assert_embed_equals(
        response, equals, attributes_to_prove=attributes_to_check
    )


async def assert_reply_matches(self, contents: str, regex):
    """ Send a message and wait for a response. If the response does not match a regex, fail the test.

    Requires a properly formatted Python regex ready to be used in the ``re`` functions.

    :param str contents: The content of the trigger message. (A command)
    :param str regex: The regular expression to test against.
    :returns: The reply.
    :rtype: discord.Message
    :raises: ResponseDidNotMatchError
    """
    response = await self.wait_for_reply(contents)
    return await self.assert_message_matches(response, regex)


async def assert_reply_has_image(self, contents):
    """Send a message consisting of `contents` and wait for a reply.

    Check that the reply contains a ``discord.Attachment``. If not, fail the test.

    :param str contents: The content of the trigger message. (A command)
    :returns: The reply.
    :rtype: discord.Message
    :raises: ResponseDidNotMatchError, NoResponseError
    """
    message = await self.wait_for_reply(contents)
    await sleep(1)  # Give discord a moment to add the embed if its a link
    return await self.assert_message_has_image(message)


async def get_delayed_reply(self, seconds_to_wait, test_function, *args):
    """Get the last reply after a specific time and check it against a given test.

    :param float seconds_to_wait: Time to wait in s
    :param method test_function: The function to call afterwards, without parenthesis
        (assert_message_equals, not assert_message_equals()!)
    :param args: The arguments to pass to the test, requires the same number of args as the test function.
        Make sur to pass in **all** args, including kwargs with defaults.
        NOTE: this policy may change if it becomes kinda stupid down the road.
    :rtype: Method
    :raises SyntaxError:
    :returns: The instance of the test requested
    """

    def parse_parameters(method) -> list:
        kwarg, parg, either, var_kwarg, var_parg = 0, 0, 0, 0, 0
        for i in signature(method).parameters:
            j = signature(method).parameters[i]
            if j.kind == _ParameterKind.KEYWORD_ONLY:
                kwarg += 1
            if j.kind == _ParameterKind.POSITIONAL_ONLY:
                parg += 1
            if j.kind == _ParameterKind.POSITIONAL_OR_KEYWORD:
                either += 1
            if j.kind == _ParameterKind.VAR_KEYWORD:
                var_kwarg += 1
            if j.kind == _ParameterKind.VAR_POSITIONAL:
                var_parg += 1
        return [kwarg, parg, either, var_kwarg, var_parg]

    desired_method_parameters = parse_parameters(test_function)
    num_desired_parameters = sum(desired_method_parameters[0:3]) - 1
    if len(args) != num_desired_parameters:
        raise SyntaxError("Invalid Number of Arguments")

    await sleep(seconds_to_wait)
    message: Message = self.channel.last_message
    return await test_function(message, *args)
